/* -------------------------------------------------------------------------
 * The Q Public License Version 1.0
 *
 *  Copyright (C) 1999 Trolltech AS, Norway.
 *  Everyone is permitted to copy and distribute this license document.
 *
 *  The intent of this license is to establish freedom to share and change
 *  the software regulated by this license under the open source model.
 *
 *  This license applies to any software containing a notice placed by the
 *  copyright holder saying that it may be distributed under the terms of
 *  the Q Public License version 1.0. Such software is herein referred to
 *  as the Software. This license covers modification and distribution of
 *  the Software, use of third-party application programs based on the
 *  Software, and development of free software which uses the Software.
 *
 *                              Granted Rights
 *
 *  1. You are granted the non-exclusive rights set forth in this license
 *  provided you agree to and comply with any and all conditions in this
 *  license. Whole or partial distribution of the Software, or software
 *  items that link with the Software, in any form signifies acceptance of
 *  this license.
 *
 *  2. You may copy and distribute the Software in unmodified form
 *  provided that the entire package, including - but not restricted to -
 *  copyright, trademark notices and disclaimers, as released by the
 *  initial developer of the Software, is distributed.
 *
 *  3. You may make modifications to the Software and distribute your
 *  modifications, in a form that is separate from the Software, such as
 *  patches. The following restrictions apply to modifications:
 *
 *    a. Modifications must not alter or remove any copyright notices in
 *    the Software.
 *
 *    b. When modifications to the Software are released under this
 *    license, a non-exclusive royalty-free right is granted to the
 *    initial developer of the Software to distribute your modification
 *    in future versions of the Software provided such versions remain
 *    available under these terms in addition to any other license(s) of
 *    the initial developer.
 *
 *  4. You may distribute machine-executable forms of the Software or
 *  machine-executable forms of modified versions of the Software,
 *  provided that you meet these restrictions:
 *
 *    a. You must include this license document in the distribution.
 *
 *    b. You must ensure that all recipients of the machine-executable
 *    forms are also able to receive the complete machine-readable source
 *    code to the distributed Software, including all modifications,
 *    without any charge beyond the costs of data transfer, and place
 *    prominent notices in the distribution explaining this.
 *
 *    c. You must ensure that all modifications included in the
 *    machine-executable forms are available under the terms of this
 *    license.
 *
 *  5. You may use the original or modified versions of the Software to
 *  compile, link and run application programs legally developed by you or
 *  by others.
 *
 *  6. You may develop application programs, reusable components and other
 *  software items that link with the original or modified versions of the
 *  Software. These items, when distributed, are subject to the following
 *  requirements:
 *
 *    a. You must ensure that all recipients of machine-executable forms
 *    of these items are also able to receive and use the complete
 *    machine-readable source code to the items without any charge beyond
 *    the costs of data transfer.
 *
 *    b. You must explicitly license all recipients of your items to use
 *    and re-distribute original and modified versions of the items in
 *    both machine-executable and source code forms. The recipients must
 *    be able to do so without any charges whatsoever, and they must be
 *    able to re-distribute to anyone they choose.
 *
 *
 *    c. If the items are not available to the general public, and the
 *    initial developer of the Software requests a copy of the items,
 *    then you must supply one.
 *
 *                         Limitations of Liability
 *
 *  In no event shall the initial developers or copyright holders be
 *  liable for any damages whatsoever, including - but not restricted to -
 *  lost revenue or profits or other direct, indirect, special, incidental
 *  or consequential damages, even if they have been advised of the
 *  possibility of such damages, except to the extent invariable law, if
 *  any, provides otherwise.
 *
 *                               No Warranty
 *
 *  The Software and this license document are provided AS IS with NO
 *  WARRANTY OF ANY KIND, INCLUDING THE WARRANTY OF DESIGN,
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *                              Choice of Law
 *
 *  This license is governed by the Laws of Norway. Disputes shall be
 *  settled by Oslo City Court.
 * -------------------------------------------------------------------------
 */

/** \mainpage Mod WASap Documentation
 *
 * \section intro Introduction
 * <PRE>
 * WASap is an Apache module (version 1.3.x), which acts as an Application level Firewall. It can filter and block malignant requests.
 *
 * The project's current stage: BETA.
 *
 * Apache Versions supported by WASap:
 *
 *   * Apache 1.3.28+ (not tested)
 *   * Apache 1.3.27 (tested)
 *   * Apache 1.3.26 (tested)
 *   * Apache 1.3.24 (tested)
 *   * Apache 1.3.12+ (not tested)
 * </PRE>
 */


#include <httpd.h>
#include <http_core.h>
#include <http_config.h>
#include <http_protocol.h>
#include <http_main.h>
#include <http_log.h>
#include <buff.h>

#include <stdlib.h>
#include <string.h>

#include <errno.h>

#include "config.h"
#include "wasap.h"

/** WASap Module Structure
 */
struct wasap_ext {
	/// module check function
	char *(*check) (request_rec *, void *, const char *,const char *, const char *);
	
	/// module configuration function
	void *(*cfg) (pool *p, void *, char *);
	
	/// module name
	char *name;
	
	/// next module in linked list
	struct wasap_ext *next;
};


struct wasap_ext *wasap_loadmods(pool *p);

int wasap_handler(request_rec *);

static const char *wasap_cfg_set_methods(cmd_parms *, wasap_dircfg *, char *);

static const char *wasap_cfg_enable(cmd_parms *, void *, char *);

static const char *wasap_cfg_set_maxgetval(cmd_parms *,wasap_dircfg *,char *);
static const char *wasap_cfg_set_maxpostval(cmd_parms *,wasap_dircfg *,char *);

static const char *wasap_cfg_set_showver(cmd_parms *, void *, int);
static const char *wasap_cfg_set_errdoc(cmd_parms *, void *, char *);

static const char *wasap_cfg_setchk(cmd_parms *, void *, char *);
static const char *wasap_cfg_addchk(cmd_parms *, void *, char *);
static const char *wasap_cfg_remchk(cmd_parms *, void *, char *);

static const char *wasap_loadmodules(cmd_parms *, void *, char *);

static const char *wasap_cfg_mod(cmd_parms *, void *, char *);

static void wasap_init(server_rec *, pool *);

static void *wasap_server_config(pool *, server_rec *);
static void *wasap_dir_cfg(pool *, char *);
static void *wasap_merge_dircfg(pool *, void *, void *);

static const char *wasap_cfg_alarm(cmd_parms *, wasap_dircfg *, char *);

void exec_alarm(request_rec *, char *);


//struct wasap_ext *wasap_exts = NULL;

/* EVERYTHING!@# */

handler_rec wasap_handlers[] = {
	{ "text/html", wasap_handler },
	{ "text/plain", wasap_handler },
	{ CGI_MAGIC_TYPE, wasap_handler },
	{ "cgi-script", wasap_handler },
	{ "application/x-httpd-php", wasap_handler},
	{ "application/x-httpd-cgi", wasap_handler },
	{ "application/x-httpd-pl", wasap_handler },
	{ "*/*", wasap_handler },
	{ NULL }
};

/** \page wasap_cmd WASap Configuration Commands
 * \section GLOBAL_SERV Server Commands
 * \subsection WASapShowVer
 *		WASapShowVer [On|Off] <BR>
 *			Sets wheter or not to show mod_wasap-(VERSION)
 *			In the Server header.
 *
 * \subsection WASapModLoad
 *		WASapModLoad / <BR>
 *			Load WASap Modules. Must be run.
 *
 * \section GLOBAL_SERVDIR Server and Directory Commands
 * \subsection WASap
 *	WASap [On|Off] <BR>
 *		Turn WASap protection on or off.
 *		Can be done at both the server and directory level.
 *
 *
 * \subsection WASapErrDoc
 *		WASapErrorDoc [Error String] <BR>
 *			Sets the content to show a client when an intrusion
 *			attempt has been spotted.
 *
 * \subsection WASapAlarm
 *		WASapAlarm [cmd] <BR>
 *			Sets a command to run when an intrusion attempt has
 *			been spotted.
 *
 * \subsection WASapMaxGetVal
 *	WASapMaxGetVal [length] <BR>
 *		Set the maximum length for a GET value.
 *
 * \subsection WASapMaxPostVal
 * 	WASapMaxPostVal [length] <BR>
 *		Set the maximum length for a POST value.
 * 
 * \subsection WASapAllowMethods
 *	WASapAllowMethods [Get,Post,...] <BR>
 *		Set allowed methods. Putting a ! before the method name will
 *		disallow it. <BR>
 *		ie: WASapAllowMethods Post,!Get <BR>
 *		Will allow POST but disallow GET.
 *
 * \subsection WASapSetChk
 *	WASapSetChk [modname0,modname1,...] <BR>
 *		Set the list of modules to run.
 *
 * \subsection WASapAddChk
 *	WASapAddChk [modname] <BR>
 *		Add a check from the current list of modules to
 *		run.
 *
 * \subsection WASapRemChk
 *	WASapRemChk [modname] <BR>
 *		Remove a check from the current list of modules to
 *		run.
 *
 * \subsection WASapMod
 *	WASapMod [modname] [config line ...] <BR>
 *		Pass config line to module modname.
 *
 */

/** \page wasap_examples WASap Examples
 * <PRE>
 * Allow GET,POST maximum length 256 a piece in / but allow PUT
 * in /dav:
 * 
 * WASap On
 * WASapModLoad /
 * WASapSetChk fmtstr,xss,file
 * 
 * &lt;Location /&gt;
 * 	WASapMaxPostVal 256
 *	WASapMaxGetVal 256
 *	
 *	WASapAllowMethods Get,Post
 * &lt;/Location&gt;
 * 
 * &lt;Location /dav&gt;
 *	WASapAllowMethods Get,Put
 * &lt;/Location&gt;
 * 
 * Try to protect PHPNuke From SQL injection and XSS
 *
 * WASap On
 * WASapModLoad /
 *
 * WASapSetChk xss,class
 *
 * WASapMod class alpha
 * WASapMod class numeric
 * WASapMod class length 256
 * WASapMod class meta=?/*!|-_.,:;
 * </PRE>
 */


static command_rec wasap_cmds[] = {
	{ "WASapAllowMethods", wasap_cfg_set_methods, NULL,
	   OR_OPTIONS, TAKE1, "define the allowed methods for WASap" },
	
	{  "WASap", wasap_cfg_enable, NULL,
	   OR_OPTIONS, TAKE1, "enable or disable WASap" },
	
	{  "WASapMaxGetVal", wasap_cfg_set_maxgetval, NULL,
	   OR_OPTIONS, TAKE1,
	   "define the maximum allowed size for a get variable" },
	
	{  "WASapMaxPostVal", wasap_cfg_set_maxpostval, NULL,
	   OR_OPTIONS, TAKE1,
	   "define the maximum allowed size for a post variable" },
	
	{  "WASapShowVer", wasap_cfg_set_showver, NULL,
	   RSRC_CONF, FLAG, "show mod_wasap version in Server: header" },
	
	{  "WASapErrDoc", wasap_cfg_set_errdoc, NULL,
	   OR_OPTIONS, TAKE1, "set mod_wasap errodc" },
	
	{ "WASapAddChk", wasap_cfg_addchk, NULL,
	   OR_OPTIONS, TAKE1, "add a check to the server, or directory" },
	
	{ "WASapRemChk", wasap_cfg_remchk, NULL,
	   OR_OPTIONS, TAKE1, "remove a check to the server or directory" },
	
	{ "WASapSetChk", wasap_cfg_setchk, NULL,
	   OR_OPTIONS, TAKE1, "set checks for the server or directory" },
	
	{ "WASapMod", wasap_cfg_mod, NULL,
	   OR_OPTIONS, RAW_ARGS, "configures modules" },
	
	{ "WASapModLoad", wasap_loadmodules, NULL,
	   OR_OPTIONS, TAKE1, "load modules" },
	
	{ "WASapAlarm", wasap_cfg_alarm, NULL,
	   OR_OPTIONS, TAKE1, "define an external alarm command for WASap" },
	
	{  NULL  }
};


module MODULE_VAR_EXPORT wasap_module;

module MODULE_VAR_EXPORT wasap_module = {
	STANDARD_MODULE_STUFF,
	wasap_init,		/* module init */
	wasap_dir_cfg,		/* module create_dir_config */
	wasap_merge_dircfg,	/* module merge_dir_config */
	wasap_server_config,	/* module create_server_config */
	NULL,
	wasap_cmds,		/* module cmds */
	wasap_handlers,		/* module handlers */
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,		   /* logger */
	NULL,		   /* header parser */
	NULL,
	NULL,
	NULL,
	NULL,
};


static void *wasap_server_config(pool * p, server_rec * rec)
{
	wasap_srvcfg *cfg = ap_pcalloc(p, sizeof(wasap_srvcfg));
	
	if (!cfg)
		return (NULL);
	
	memset(cfg, 0, sizeof(wasap_srvcfg));
	
	cfg->methods = ((1 << M_POST) | (1 << M_GET));
	cfg->on = WSPON; // default is to turn on (for now);
	cfg->maxver = MAX_VER;
	cfg->showver = TRUE;
	cfg->maxgetval = MAX_GETVAL;
	cfg->maxheader = MAX_HEADER;
	cfg->maxpostval = MAX_POSTVAL;
	
	cfg->checks = NULL;
	cfg->modcfg = ap_make_table(p,8);
	
	cfg->errdoc = ap_pstrdup(p,
			"<html>"
			"<title>Mod WASap</title>"
			"<body>Logging Illegal HTTP Request</body>"
			"</html>"
		);
	
	return ((void *) cfg);
	
}


static void *wasap_dir_cfg(pool * p, char *dir)
{
	wasap_dircfg *cfg = ap_pcalloc(p, sizeof(wasap_dircfg));
	
	
	if (cfg == NULL)
		return (NULL);
	
	memset(cfg, 0, sizeof(wasap_dircfg));
	
	cfg->on = WSPNONE;      // so we can check if we switch on/off
	cfg->dir = ap_pstrdup(p, dir);
	
	cfg->checks = NULL;
	cfg->modcfg = ap_make_table(p,8);
	
	return ((void *) cfg);
	
}


static void *wasap_merge_dircfg(pool * p, void *parent, void *subdir)
{
	wasap_dircfg *pcfg = (wasap_dircfg *)parent;
	wasap_dircfg *scfg = (wasap_dircfg *)subdir;
	wasap_dircfg *new = ap_pcalloc(p, sizeof(wasap_dircfg));
	
	
	if (pcfg->dir) // && pcfg->on==WSPON) // ignore empty and disabled directories
		memcpy(new, pcfg, sizeof(wasap_dircfg));
	else {
		memcpy(new, scfg, sizeof(wasap_dircfg));
		return (new);
	}
	
	if (scfg->on != WSPNONE)
		new->on = scfg->on;
	
	if (scfg->methods)
		new->methods = scfg->methods;
	
	if (scfg->maxgetval)
		new->maxgetval = scfg->maxgetval;
	
	if (scfg->maxpostval)
		new->maxpostval = scfg->maxpostval;
	
	if (scfg->checks)
		new->checks = scfg->checks;
	
	if (scfg->modcfg)
		new->modcfg = scfg->modcfg;
	
	return ((void *) new);
	
}


static const char *wasap_cfg_set_showver(cmd_parms * cmd, void *dummy,
					 int arg)
{
	server_rec *s = cmd->server;
	wasap_srvcfg *srvcfg =
	    ap_get_module_config(s->module_config, &wasap_module);
	
	srvcfg->showver = arg;
	
	return (NULL);
}


static const char *wasap_cfg_set_errdoc(cmd_parms * cmd, void *dummy,
					char *arg)
{
	server_rec *s = cmd->server;
	wasap_srvcfg *srvcfg =
		ap_get_module_config(s->module_config, &wasap_module);
	
	srvcfg->errdoc = ap_pstrdup(cmd->pool, arg);
	
	return (NULL);
}


static const char *wasap_cfg_set_methods(cmd_parms * cmd,
					 wasap_dircfg * dircfg, char *arg)
{
	server_rec *s = cmd->server;
	char *ptr, *token, *buf, *p;
	int method, methods, neg = 0;
	wasap_srvcfg *srvcfg =
	    ap_get_module_config(s->module_config, &wasap_module);
	
	
	buf = ap_pstrdup(cmd->temp_pool, arg);
	
	if (cmd->path)
		methods = dircfg->methods;
	else
		methods = srvcfg->methods;
	
	for (ptr = strtok_r(buf, ",", &token); ptr;
				ptr = strtok_r(NULL, ",", &token)) {
			
		if (ptr[0] == '*') {
			methods = ~M_INVALID;
			continue;
		}
		
		if (ptr[0] == '!') {
			neg = 1;
			ptr++;
		}
		
		for (p = ptr; *p; p++)
			*p = toupper(*p);
		
		if ((method = ap_method_number_of(ptr)) >= M_INVALID) {
			
			buf = ap_palloc(cmd->temp_pool, 128);
			ap_snprintf(buf, 128,
				    "mod_wasap: illegal method %s",ptr);
			
			return (buf);
		}
		
		if (!neg)
			methods |= (1 << method);
		else {
			methods &= ~(1 << method);
			neg = 0;
		}
		
	}
	
	
	if (cmd->path)
		dircfg->methods = methods;
	else
		srvcfg->methods = methods;

	return (NULL);
	
}


static const char *wasap_cfg_set_maxgetval(cmd_parms * cmd,
					   wasap_dircfg * dircfg,
					   char *arg)
{
	server_rec *s = cmd->server;
	char *ptr, *token, *buf, *p;
	int method, methods, neg = 0;
	wasap_srvcfg *srvcfg =
	    ap_get_module_config(s->module_config, &wasap_module);
	
	
	if (cmd->path)
		methods = dircfg->maxgetval = atoi(arg);
	else
		methods = srvcfg->maxgetval = atoi(arg);
	
	
	return (NULL);
}


static const char *wasap_cfg_set_maxpostval(cmd_parms * cmd,
					    wasap_dircfg * dircfg,
					    char *arg)
{
	server_rec *s = cmd->server;
	char *ptr, *token, *buf, *p;
	int method, methods, neg = 0;
	wasap_srvcfg *srvcfg =
	    ap_get_module_config(s->module_config, &wasap_module);
	
	
	if (cmd->path)
		methods = dircfg->maxpostval = atoi(arg);
	else
		methods = srvcfg->maxpostval = atoi(arg);
	
	return (NULL);
}


static const char *wasap_cfg_enable(cmd_parms * cmd, void *dummy,
				    char *arg)
{
	int set;
	server_rec *s = cmd->server;
	wasap_srvcfg *srvcfg =
		ap_get_module_config(s->module_config, &wasap_module);
	wasap_dircfg *dircfg = (wasap_dircfg *) dummy;
	
	
	if (!strcasecmp("Off", arg))
		set = WSPOFF;
	else if (!strcasecmp("On", arg))
		set = WSPON;
	else
		return (ap_pstrdup
			(cmd->pool, "Invalid arg for WASap Directive"));
	
	if (cmd->path)
		dircfg->on = set;
	else
		srvcfg->on = set;
	
	return (NULL);
}


static const char *wasap_cfg_setchk(cmd_parms * cmd, void *dummy,
				    char *arg)
{
	server_rec *s = cmd->server;
	wasap_srvcfg *srvcfg =
		ap_get_module_config(s->module_config, &wasap_module);
	wasap_dircfg *dircfg = (wasap_dircfg *) dummy;
	char *buf,*ptr,*token,*p;
	int neg=0;
	table *checks;
	
	buf = ap_pstrdup(cmd->temp_pool, arg);
	
	if (cmd->path) {
		if (dircfg->checks == NULL)
			dircfg->checks = ap_make_table(cmd->pool,8);
		
		checks = dircfg->checks;
	} else {
		if (srvcfg->checks == NULL)
			srvcfg->checks = ap_make_table(cmd->pool,8);
		
		checks = srvcfg->checks;
	}
	
	
	for (ptr = strtok_r(buf, ",", &token); ptr;
	     ptr = strtok_r(NULL, ",", &token)) {
		
		
		if (ptr[0] == '!') {
			neg = 1;
			ptr++;
		}
		
		for (p = ptr; *p; p++)
			*p = tolower(*p);
		
		if (neg) {
			neg=0; 
			ap_table_setn(checks,ptr,"off");
		}
		else
			ap_table_setn(checks,ptr,"on");
	}

	
	return (NULL);
}


static const char *wasap_loadmodules(cmd_parms *cmd, void *dummy, char *arg)
{
	server_rec *s = cmd->server;
	wasap_srvcfg *srvcfg =
		ap_get_module_config(s->module_config, &wasap_module);
	
	srvcfg->wasap_exts = wasap_loadmods(cmd->pool);
	
	return(NULL);
}

static const char *wasap_cfg_addchk(cmd_parms * cmd, void *dummy,
				    char *arg)
{
	server_rec *s = cmd->server;
	wasap_srvcfg *srvcfg =
		ap_get_module_config(s->module_config, &wasap_module);
	wasap_dircfg *dircfg = (wasap_dircfg *) dummy;
	
	
	if (cmd->path) {
		if (dircfg->checks == NULL)
			dircfg->checks = ap_make_table(cmd->pool,8);
		
		ap_table_setn(dircfg->checks,arg,"on");
	} else {
		if (srvcfg->checks == NULL)
			srvcfg->checks = ap_make_table(cmd->pool,8);
		
		ap_table_setn(srvcfg->checks,arg,"on");
	}
	
	return (NULL);
}


static const char *wasap_cfg_remchk(cmd_parms * cmd, void *dummy,
				    char *arg)
{
	server_rec *s = cmd->server;
	wasap_srvcfg *srvcfg =
		ap_get_module_config(s->module_config, &wasap_module);
	wasap_dircfg *dircfg = (wasap_dircfg *) dummy;
	
	
	if (cmd->path) {
		if (dircfg->checks == NULL)
			dircfg->checks = ap_make_table(cmd->pool,8);
		
		ap_table_setn(dircfg->checks,arg,"off");
	} else {
		if (srvcfg->checks == NULL)
			srvcfg->checks = ap_make_table(cmd->pool,8);
		
		ap_table_setn(srvcfg->checks,arg,"off");
	}
	
	return (NULL);
}


static const char *wasap_cfg_mod(cmd_parms *cmd, void *dummy, char *args)
{
	server_rec *s = cmd->server;
	char *modname,*arg,*tmp;
	struct wasap_ext *mod;
	table *modcfg;
	void *cfg;
	
	wasap_srvcfg *srvcfg = 
		ap_get_module_config(s->module_config, &wasap_module);
	wasap_dircfg *dircfg = (wasap_dircfg *) dummy;
	
	
	if (srvcfg->wasap_exts == NULL)
		return(NULL);
	
	
	if (cmd->path) {
		if (dircfg->modcfg == NULL)
			dircfg->modcfg = ap_make_table(cmd->pool,8);
		
		modcfg = dircfg->modcfg;
	} else {
		if (srvcfg->modcfg == NULL)
			srvcfg->modcfg = ap_make_table(cmd->pool,8);
		
		modcfg = srvcfg->modcfg;
	}
	
	if ((tmp = ap_pstrdup(cmd->temp_pool,args)) == NULL)
		return("ENOMEM");
	
	modname = tmp;
	
	ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,NULL,
			"module %s is to be config'd",modname);
		
	if ((arg = strchr(modname,' ')) == NULL)
		return("Syntax Error");
	
	*arg++ = '\0';
	
	for (mod = srvcfg->wasap_exts; mod ; mod = mod->next)
		if (!strcmp(mod->name,modname))
			break;
	
	if (!mod) {
		tmp = ap_psprintf(cmd->temp_pool,"no module: %s",modname);
		return(tmp);
	}
	
	if (!mod->cfg)
		return(NULL);
	
	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, NULL,
			"module %s to be passed %s",modname,arg);
	
	cfg = (void *)ap_table_get(modcfg,mod->name);
	cfg = (void *)mod->cfg(cmd->pool,cfg,arg);
	ap_table_setn(modcfg,mod->name,cfg);	
	
	
	return(NULL);
	
}


static const char *wasap_cfg_alarm(cmd_parms * cmd,
					 wasap_dircfg * dircfg, char *arg)
{
	server_rec *s = cmd->server;
	char *ptr, *token, *buf, *p;
	int method, methods, neg = 0;
	wasap_srvcfg *srvcfg =
	    ap_get_module_config(s->module_config, &wasap_module);
	
	buf = ap_pstrdup(cmd->pool, arg);
	
	if (cmd->path)
		dircfg->alarm_cfg = buf;
	else
		srvcfg->alarm_cfg = buf;
	
	return(NULL);
}



void exec_alarm(request_rec *r, char *alert)
{
	char *cmd;
	char exec[4096];
	char *args[4];
	wasap_srvcfg *scfg =
	    (wasap_srvcfg *) ap_get_module_config(r->server->module_config,
						  &wasap_module);
	wasap_dircfg *dcfg =
	    (wasap_dircfg *) ap_get_module_config(r->per_dir_config,
						  &wasap_module);
	
	if (!(cmd=dcfg->alarm_cfg))
		if (!(cmd=scfg->alarm_cfg))
			return;
	
	snprintf(exec,sizeof(exec),"%s %s \"%s\"",
				cmd,
				r->connection->remote_ip,
				ap_escape_quotes(r->pool,alert));
	
	ap_log_error(APLOG_MARK, APLOG_NOTICE|APLOG_NOERRNO,NULL,
			"log alert: %s",exec);
	
	args[0] = "/bin/bash"; /* should be ./configure'able ... */
	args[1] = "-c";
	args[2] = exec;
	args[3] = NULL;
	
	
	/* switch to vfork() ? */
	/* if we dont wait() or waitpid() will we leave zombies? */
	
	switch(fork()) {
	case 0:
		execve(args[0],args,NULL);
		exit(0);
	break;
	case -1:
		//ap_log_error(
	break;
	}
	
}

char *exec_check(
			struct wasap_ext *chk,
			request_rec *r,
			int header,
			const char *file,
			const char *key,
			const char *val
		)
{
	char *tmp;
	int dchk=0,schk=0;
	void *modcfg=NULL;
	char *ret;
	char err[4096];
		
	wasap_srvcfg *scfg =
	    (wasap_srvcfg *) ap_get_module_config(r->server->module_config,
						  &wasap_module);
	wasap_dircfg *dcfg =
	    (wasap_dircfg *) ap_get_module_config(r->per_dir_config,
						  &wasap_module);
	
	
	if ( dcfg && !header ) {
		
		if ( dcfg->checks ) {
			tmp = (char *)ap_table_get(dcfg->checks,chk->name);
			
			if (!tmp || strcmp(tmp,"on") == 0) {
				ap_log_error(APLOG_MARK,
						APLOG_DEBUG|APLOG_NOERRNO,NULL,
						"module: %s is disactivated",
						chk->name
					);
				
				dchk=1;
			}
		}
		
		if ( dcfg->modcfg )
			modcfg = (void *)ap_table_get(dcfg->modcfg,chk->name);
		
	}
	
	
	if ( scfg ) {
		if ( scfg->checks ) {
			tmp = (char *)ap_table_get(scfg->checks,chk->name);
			
			if (!tmp || strcmp(tmp,"on") == 0) {
				ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,NULL,
				"module: %s is deactivaed with %s",chk->name,tmp);
				
				schk=1;
			}
		}
		
		if ( !modcfg && scfg->modcfg )
			modcfg = (void *)ap_table_get(scfg->modcfg,chk->name);
		
	}
	
	
	if (dchk || schk) {
		ret=chk->check(r,modcfg,file,key,val);
		if ( ret ) {
			snprintf(err,sizeof(err),"%s in %s?%s=%s",
							ret,file,key,val);
			
			exec_alarm(r,err);
			return(ret);
		}
	} else
		return(NULL);
	
	
	return(NULL); /* NEVER REACHED */
	
}

int check_args(request_rec * r, const char *args)
{
	const char *ptr;
	char *buf, *key, *val, *token, *err;
	const char *chk;
	int len, lcmp, maxgetval, maxpostval;
	struct wasap_ext *cur;
	void *modcfg;
	
	wasap_srvcfg *scfg =
	    (wasap_srvcfg *) ap_get_module_config(r->server->module_config,
						  &wasap_module);
	wasap_dircfg *dcfg =
	    (wasap_dircfg *) ap_get_module_config(r->per_dir_config,
						  &wasap_module);
	
		
	ptr = (args) ? args : r->args;
	if (ptr == NULL) 
		return (TRUE);
		
	/* this should be changed */
	
	if ((buf = ap_pstrdup(r->pool, ptr)) == NULL)
		return (TRUE);
	
	
	if (dcfg->maxgetval > 0 && dcfg->dir) {
		maxgetval = dcfg->maxgetval;
		maxpostval = dcfg->maxpostval;
	} else {
		maxgetval = scfg->maxgetval;
		maxpostval = scfg->maxpostval;
	}
	
	
	switch (r->method_number) {
	case M_GET:
		lcmp = maxgetval;
		break;
	case M_POST:
		lcmp = maxpostval;
		break;
	}

	
	for (ptr = strtok_r(buf, "&", &token); ptr;
		ptr = strtok_r(NULL, "&", &token)) {
		
		ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,NULL,
				"checking %s",ptr);
		
		if ((val = strchr(ptr, '=')) == NULL)
			continue;
		*val++ = (char) 0;

		ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,NULL,
				"checking var=%s val=%s",ptr,val);
		
		key = trans_uri(r->pool, ptr, TRANS_CHAR);
		val = trans_uri(r->pool, val, TRANS_CHAR|TRANS_PATH);
		
		
		if (val == NULL || key == NULL) {
			ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,NULL,
					"ran out of memory in check_args()");
			return(FALSE);
		}

		ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,NULL,
				"checking (trans'd) var=%s val=%s",ptr,val);
		
		len = strlen(val);
		
		if (len > lcmp) {
			ap_log_rerror(APLOG_MARK, APLOG_CRIT, r,
				      "mod_wasap: large value, size=%d\n",
				      len);
			return (FALSE);
		}
		
		
		for (cur = scfg->wasap_exts; cur; cur = cur->next) {
			ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO,NULL,
					"module: %s",cur->name);
			
			err = exec_check(cur,r,FALSE,r->filename,key,val);
			
			if (err != NULL) {
				ap_log_rerror(APLOG_MARK, APLOG_CRIT|APLOG_NOERRNO,r,
						"mod_wasap[%s]: %s in %s=%s against %s",
						ap_escape_logitem(r->pool,cur->name),
						ap_escape_logitem(r->pool,err),
						ap_escape_logitem(r->pool,key),
						ap_escape_logitem(r->pool,val),
						ap_escape_logitem(r->pool,r->filename)
					);
				return (FALSE);
			}
			
		}
		
	}
	
	return (TRUE);
	
}


int check_method(request_rec * r, const char *method)
{
	char *end, *ver, *req;
	int major, minor, methods, i;
	
	wasap_srvcfg *srvcfg =
	    (wasap_srvcfg *) ap_get_module_config(r->server->module_config,
						  &wasap_module);
	wasap_dircfg *dircfg =
	    (wasap_dircfg *) ap_get_module_config(r->per_dir_config,
						  &wasap_module);
	
	if (dircfg->methods)
		methods = dircfg->methods;
	else
		methods = srvcfg->methods;

	if ((1 << r->method_number) & methods)
		return (TRUE);

	ap_log_rerror(APLOG_MARK, APLOG_CRIT, r,
		      "mod_wasap: illegal method: %s for %s", 
		      ap_escape_logitem(r->pool,r->method),
		      ap_escape_logitem(r->pool,r->uri));
	
	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r,
		      "mod_wasap: method_number=%d methods=%d",
		      r->method_number, methods);
	
	exec_alarm(r,"illegal method");
	return (FALSE);	 /* 405 */
	
}


int check_headers(void *req, const char *okey, const char *oval)
{
	int i;
	struct wasap_ext *cur;
	char *err;
	char buf[64];
	void *modcfg;
	char *key,*val;
	
	wasap_rec *wr = (wasap_rec *) req;
	
	
	if (strlen(oval) > wr->scfg->maxheader) {
		
		ap_log_rerror(APLOG_MARK,APLOG_CRIT,wr->rec,
			"header: %s is too long", okey);
		
		wr->tmp = FALSE;
		return (FALSE);
		
	}
	
	
	key=trans_uri(wr->pool,okey,TRANS_CHAR);
	val=trans_uri(wr->pool,oval,TRANS_CHAR);
	
	if ((!key) || (!val))
		return (FALSE);
	
	for (cur = wr->scfg->wasap_exts; cur; cur = cur->next) {
		snprintf(buf, sizeof(buf), "header:%s", key);
		
		err = exec_check(cur,wr->rec,TRUE,buf,key,val);
		if (err != NULL) {
			
			ap_log_rerror(APLOG_MARK, APLOG_CRIT,wr->rec,
					"mod_wasap[%s]: %s in %s=%s", 
					cur->name,err,buf,val);
			
			wr->tmp = FALSE;
			return (FALSE);
			
		}
		
	}
	
	return(TRUE);
}

/** \fn void wasap_reinject(request_rec *r, void *buf, int nbyte )
 * \brief reinject preread information back into the apache buffer.
 * \param r request_rec request_rec of the buffer to inject.
 * \param buf data to inject into the buffer
 * \param nbyte number of bytes to inject
 *
 */

static void wasap_reinject(request_rec * r, void *buf, int nbyte)
{
	BUFF *fb;
	char *p;
	
	fb = r->connection->client;
	fb->inptr = fb->inbase = buf;   /* restart, w00h00 */
	fb->bufsiz = nbyte;
	fb->incnt = nbyte;
	
}


static char *wasap_readbody(request_rec *r)
{
	char *ptr, *new,*payload;
	char buf[8192];
	unsigned long remaining, read_length;
	unsigned long len_buf = 0, bufsiz = sizeof(buf);
	unsigned long len_read;
	
	
	ap_hard_timeout("mod_wasap input buffering", r);
	
	if ((ptr = payload = malloc(bufsiz + 1)) == NULL) {
		ap_log_error(APLOG_MARK,
			     APLOG_CRIT | APLOG_NOERRNO, NULL,
			     "mod_wasap: ran out of memory");
		ap_die(-1);
	}
	
	/* We same r->remaing and r->read_length so we can
	   later rewind Apache's internal read buffer so
	   mod_php, mod_perl and others can read the buffer
	   later. (this is a hack, so it has to be fully commented).
	   
	   This routine DOES NOT use apache's pool memory
	   because we could cause a spiral effect of memory
	   allocations which are NEVER free()'d or deallocated.
	*/
	
	remaining = r->remaining;
	read_length = r->read_length;
	
	ap_setup_client_block(r, REQUEST_CHUNKED_ERROR);
	
	
	/* Read all the input for later inspection */
	
	while ((len_read = ap_get_client_block(r, buf, sizeof(buf))) > 0) {
		
		ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,NULL,
			     "mod_wasap: chunk[%d], bufsiz=%d", len_read,
			     bufsiz);
		
		/* check to see if we must increase the buffer */
		
		if ((len_buf + len_read) > bufsiz) {
			
			bufsiz += sizeof(buf);
			new = realloc(payload, bufsiz + 1);
			if (new == NULL) {
				/* later we can make this more
				   graceful
				*/
				
				ap_die(-1);
			}
			
			/* advance the buffer pointer */
			
			ptr = new + (ptr - payload);
			payload = new;
			
		}
		
		/* increase the buffer size, copy into buffer
		   advance our buffer pointer */
		
		len_buf += len_read;
		memcpy(ptr, buf, len_read);
		ptr += len_read;
		
		ap_reset_timeout(r);
		
	}
	
	ap_kill_timeout(r);
	
	
	/* could this NULL be dangerous */
	
	/* here we put a NULL at the end of the buffer, just
	   for convenience. We also consolidate the buffer
	   back into apache's pool space.
	*/
	
	payload[len_buf] = (char) NULL;
	if ((new = ap_pstrdup(r->connection->pool, payload)) == NULL) {
		ap_log_error(APLOG_MARK,
			     APLOG_CRIT | APLOG_NOERRNO, NULL,
			     "mod_wasap: ran out of memory");
		ap_die(-1);
	}
	
	free(payload);
	payload = new;
	
	wasap_reinject(r, payload, len_buf);
	
	/* rewind the buffer */
	r->remaining = remaining;
	r->read_length = read_length;
	
	
	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, NULL,
		     "mod_wasap: done reinject [%d,%d]: \n%s\n",
		     len_read, bufsiz, payload);
	
	return(payload);
}		

int wasap_handler(request_rec * r)
{
	char *payload = NULL;
	wasap_rec *wr;
	int ret;
	
	wasap_srvcfg *srvcfg = (wasap_srvcfg *) 
		ap_get_module_config(r->server->module_config, &wasap_module);
	wasap_dircfg *dircfg = (wasap_dircfg *) 
		ap_get_module_config(r->per_dir_config, &wasap_module);
	
	
	ap_log_error(APLOG_MARK, APLOG_DEBUG, NULL,
		     "mod_wasap: entering wasap_handler");
	
		
	if (dircfg->on != WSPNONE && dircfg->dir)
		ret = dircfg->on;
	else
		ret = srvcfg->on;
	
	if (ret == WSPOFF)
		return (DECLINED);
	
	
	wr = ap_pcalloc(r->pool,sizeof(wasap_rec));
	if (wr == NULL) {
		// ap_log_error();
		ap_die(-1);
	}
	
	wr->alloc = (void *)ap_palloc;
	wr->pool = r->pool;
	wr->rec = r;
	wr->scfg = srvcfg;
	wr->dcfg = dircfg;
	
		
	/* we check the args here because GET args can still be processed
	   (at least in PHP) with a HEAD verb
	*/
	
	if ( check_args(r, NULL) == FALSE ) {
		ap_custom_response(r, HTTP_BAD_REQUEST, srvcfg->errdoc);
		return (HTTP_BAD_REQUEST);
	}
	
	
	
	wr->tmp = TRUE;
	ap_table_do(check_headers, wr, r->headers_in, NULL);
	
	if (wr->tmp == FALSE) {
		ap_custom_response(r, HTTP_BAD_REQUEST, srvcfg->errdoc);
		return (HTTP_BAD_REQUEST);
	}
	
	
	if (r->header_only)
		return (DECLINED);
	
	
	/* Buffer all input for later inspection */
		
	if (ap_should_client_block(r))
		payload = wasap_readbody(r);
	
	
	if ( 
		check_method(r,r->method) == FALSE ||
		check_proto(wr,r->protocol) == FALSE ||
		check_scheme(wr,r->the_request) == FALSE ||
		check_uri(wr,r->unparsed_uri) == FALSE
	    ) {
		ap_custom_response(r, HTTP_BAD_REQUEST, srvcfg->errdoc);
		return (HTTP_BAD_REQUEST);
	}
	
	
	/* multipart/form-data missing support */
	
	if (r->method_number == M_POST) {
		
		if (check_args(r, payload) == FALSE) {
			
			ap_custom_response(r, HTTP_BAD_REQUEST,
					   srvcfg->errdoc);
			return (HTTP_BAD_REQUEST);
		}
	}
	
	return (DECLINED);
	
}

struct wasap_ext *wasap_loadmods(pool *p);

static void wasap_init(server_rec * s, pool * p)
{
	char *dirname;
	wasap_srvcfg *cfg =
	    ap_get_module_config(s->module_config, &wasap_module);
	
	ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,NULL,
			"wasap_init executed");
	
	if (cfg) {
		if (cfg->on == WSPON && cfg->showver == TRUE)
			ap_add_version_component("WASap/" MOD_VERSION);
	}
	
	
	ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,NULL,
			"wasap_exts: 0x%x checks: 0x%x",cfg->wasap_exts,cfg->checks);
	
	/*
	if (!wasap_exts) {
		if (cfg->wasap_exts != NULL)
			wasap_exts = cfg->wasap_exts;
		else
			wasap_loadmods(p);
	}
	
	cfg->wasap_exts = wasap_exts;
	ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,NULL,
			"wasap_exts: 0x%x 0x%x",cfg->wasap_exts,wasap_exts);
	*/
}


#define WASAP_MODDIR LIBEXECDIR "/WASap/"

#ifndef WASAP_STATIC_MODULES
struct wasap_ext *wasap_loadmods(pool *p)
{
	char *tmp;
	char *(*modname)(void);
	char *(*check)(request_rec *, void *, const char *, const char *, const char *);
	void *(*cfg)(pool *p, void *, char *);
	void *handle;
	DIR *dirfd;
	struct wasap_ext *wasap_exts=NULL,*cur, *prev;
	struct dirent *dent;
	
	/* apache probablemente tiene algun ap_popendir() ... 
	   esto podria cambiar. */
	
	if ((dirfd = opendir(WASAP_MODDIR)) == NULL) {
		ap_log_error(APLOG_MARK, APLOG_CRIT | APLOG_NOERRNO, NULL,
			     "mod_wasap: unable to open modules directory %s",
			     WASAP_MODDIR);

		return;
	}
	
	prev = NULL;
	
	ap_log_error(APLOG_MARK, APLOG_CRIT | APLOG_NOERRNO, NULL,
		     "mod_wasap: loading modules");
	
	while ((dent = readdir(dirfd))) {
		if ((tmp = strrchr(dent->d_name, '.')) == NULL)
			continue;

		if (strcmp(tmp, ".so"))
			continue;
		
		tmp = ap_pstrcat(p, WASAP_MODDIR, dent->d_name, NULL);
		
		if ((handle = ap_os_dso_load(tmp)) == NULL) {
			ap_log_error(APLOG_MARK, APLOG_ERR, NULL,
				     "mod_wasap: unable to load module: %s, reason: %s",
				     tmp,dlerror());
			continue;
		}
		
		modname = ap_os_dso_sym(handle, "chk_name");
		if (!modname)
			continue;
		
		tmp = ap_pstrcat(p, "chk_", modname(), NULL);
		
		check = ap_os_dso_sym(handle,tmp);
		if (!check)
			continue;
		
		if ((cur =
		     ap_pcalloc(p, sizeof(struct wasap_ext))) == NULL) {
			ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO,
				     NULL, "mod_wasap: ran out of memory");
			ap_die(-1);
		}
						
		cur->name = modname();
		cur->check = check;
		
		
		tmp = ap_pstrcat(p, "chk_", modname(), "_cfg", NULL);
		cfg = ap_os_dso_sym(handle,tmp);
		if (cfg)
			cur->cfg = cfg;
		
		
		ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, NULL,
			     "mod_wasap: loaded module %s", dent->d_name);
		
		if (prev)
			prev->next = cur;
		else
			wasap_exts = cur;       // is this the FIRST time?
		
		prev = cur;
	}
	
	closedir(dirfd);
	return(wasap_exts);
	
}
#else
/* como lo haremos, eh?
 *
 * Algo asi como un generador de C que rellene la estructura.
 * Tambien podria ser un MACRO.
 *
 */
wasap_loadmods(void *dummy) {
	return;
}
#endif
