#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <httpd.h>

/* 
   this algorithmic table was created with a bitmask for EVERY metacharacter.
   its only flaw is that '/' is defined with the same value as '\'.
   From a security standpoint, this is a moot point, but from a functionality
   standpoint, this COULD be bad... (most likely for sloppy php coders or
   31337 asp c0derZ who use jscript tekneeqz to esc chars (<-none found))
*/

/** \page class_doc Class Check Documentation
 * \section introduction Introduction
 *	Class Check allows one to classify the expected input
 *	by allowing only characters within one of three groups;
 *	alpha, numeric, and metacharacters. The metacharacters 
 *	group is divided into the 32 metacharacters (which are all
 *	printable characters which are not alphanumeric), <P> ie: <BR>
 *	brackets ({}) , minus sign (-) , parentheses (()). </P>
 */


struct vtype {
  unsigned int meta;
  int alnum; // 1 = alpha, 2 = numeric
};


#define CLASS(a) (cmap[((a)-0x20)])
#define CLASS_ALL (0x8fffffff)
#define CLASS_BAD(a) ((a < 0x20) || (a > 0x7e))

#define CLASS_ALPHA (2)
#define CLASS_NUMERIC (1)
#define CLASS_BITS ((CLASS_ALPHA | CLASS_NUMERIC))


static signed int cmap[] = {
	0x00000000,0x00000001,0x00000002,0x00000004,0x00000008,
	0x00000010,0x00000020,0x00000040,0x00000080,0x00000100,
	0x00000200,0x00000400,0x00000800,0x00001000,0x00002000,
	0x00004000,0xffffffff,0xffffffff,0xffffffff,0xffffffff,
	0xffffffff,0xffffffff,0xffffffff,0xffffffff,0xffffffff,
	0xffffffff,0x00008000,0x00010000,0x00020000,0x00040000,
	0x00080000,0x00100000,0x00200000,0x00000000,0x00000000,
	0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,
	0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,
	0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,
	0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,
	0x00000000,0x00000000,0x00000000,0x00000000,0x00400000,
	0x00800000,0x01000000,0x02000000,0x04000000,0x08000000,
	0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,
	0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,
	0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,
	0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,
	0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,
	0x00000000,0x10000000,0x20000000,0x40000000,0x80000000
};
                                                                                                                                                        

char *chk_name(void)
{
	return("class");
}

/** \page class_doc
 *
 * \section Configuration
 *
 * the Class Check allows you to configure the character class, so to speak,
 * that is allowed for a certain area (server or directory). <BR>
 * To configure Class Check you need a Directive of the following form
 * in the Apache configuration files: <BR>
 * <P>
 * WASapMod class (!)&lt;keyword0&gt;[=value] &lt;keyword1&gt;[=value] ...<BR>
 * (the class module IS sensitive to space)
 * </P>
 * <BR>
 * <BR>
 * The prefixed ! is used to disallow those characters specificied. <BR>
 * the class check has the following keywords:
 * 
 * \subsection Alpha alpha
 *	The alpha keyword allows alphabetical characters through.
 *
 * \subsection Numeric numeric
 *	The numeric keyword allows numeric chacters through.
 * \subsection Meta meta
 *	The meta keyword allows the metacharacters in value through.
 *
 * \subsection Length length
 *	The length keyword allows values only as big as the value given.
 *
 * \section example Examples
 *
 * \subsection example1 WASapMod class meta=';-" alpha numeric length=8
 *	This rather dangerous example would allow only dangerous
 *	SQL characters, but only allow values of up to 8 bytes.
 *
 * \subsection example2 WASapMod class !meta=';-" alpha numeric 
 *	This example would allow alpha numeric strings of any length
 *	without the characters: (';-")
 *
 */

void *chk_class_cfg(pool *p, void *cfg, char *args)
{
	char *chunk,*token;
	char *key,*val;
	char *ptr;
	struct vtype *ret;
	int neg,map;
	
	
	if (cfg)
		ret = cfg;
	else {
		ret = calloc(1,sizeof(struct vtype));
		if (!ret)
			return(NULL);
		
		//ret->meta = CLASS_ALL;
		//ret->alnum = CLASS_BITS;
		
	}
	
	
	chunk = strtok_r(args," ",&token);
	
	do {
		key = chunk;
		neg = 0;
		
		if (*key == '!') {
			key++;
			neg=1;
		}
		
		if ((val = strchr(key,'=')))
			*val++ = '\0';
		
		if (strcmp(key,"meta") == 0) {
			if (!val)
				continue;
			map = 0;
			
			for (ptr = val; *ptr ; ptr++) {
				if (CLASS_BAD(*ptr))
					return(NULL);
				
				map |= (CLASS(*ptr));
			}
			
			
			if (!neg)
				ret->meta |= map;
			else
				ret->meta &= ~map;
			
		}
		else if (strcmp(key,"alpha") == 0) {
			if (!neg)
				ret->alnum |= CLASS_ALPHA;
			else
				ret->alnum &= ~CLASS_ALPHA;
		}
		else if (strcmp(key,"numeric") == 0) {
			if (!neg)
				ret->alnum |= CLASS_NUMERIC;
			else
				ret->alnum &= ~CLASS_NUMERIC;
		}
		else if (strcmp(key,"length") == 0) {
			unsigned len;
			
			if (!val)
				continue;
			
			len = atoi(val);
			
			/* erase previous size bits, just in case */
			ret->alnum &= CLASS_BITS;
			ret->alnum |= (len & ~CLASS_BITS);
		}
		
	} while((chunk = strtok_r(NULL," ",&token)));
	
	
	return(ret);
	
}


char *chk_class(request_rec *r, void *dummy,char *file,char *key,char *val)
{
	char *s;
	struct vtype *vt = (struct vtype *) dummy;
	char *buf;
	int class;
	
	
	if (vt == NULL)
		return(NULL);
	
	for ( s = val ; *s ; s++) {
		if (CLASS_BAD(*s))
			return("Illegal character for chk_class");
		
		class = CLASS(*s);
		
		switch(class) {
		case 0:
			if (!(vt->alnum & CLASS_NUMERIC))
				return("Illegal Character [numeric]");
		break;
		case -1:
			if (!(vt->alnum & CLASS_ALPHA))
				return("Illegal Character [alpha]");
		break;
		default:
			
			if (!(vt->meta & class)) {
				buf=ap_palloc(r->pool,256);
				
				snprintf(buf,256,"Illegal Character [%c]",*s);
				return(buf);
			}
			
		break;
		}
		
	}
	
	
	if ((vt->alnum & ~3) && ((vt->alnum & ~3) < (s - val)))
		return("Value is too long");
	
	return(NULL);
	
}
