#define MOD_VERSION "0.0.11"

#define MAX_MINOR 1
#define MAX_MAJOR 1
#define MAKE_VER(a,b) ((a << 8) | (b))
#define MAX_VER MAKE_VER(MAX_MAJOR,MAX_MINOR)

#define WSPON 2
#define WSPOFF 1
#define WSPNONE 0
#define WSPDEF WSPOFF


#define FALSE 0
#define TRUE 1

#define TRANS_CHAR ((1 << 0))
#define TRANS_PATH ((1 << 1))

/* Quite a lot of these are ALREADY handled by Apache */

/** WASap Server Configuration Structure
 *
 * This is the Structure which holds the Server Configuration
 * settings.
 */
typedef struct {
	/// Allowed Methods
	/// corresponds with the M_* values
	int methods;

	/// Activated or not
	int on;

	/// List of activated checks
	table *checks;
	
	/// Max URI Length
	int maxuri;

	/// Max GET value length
	int maxgetval;

	/// Max POST value length
	int maxpostval;

	/// Max Header value length
	int maxheader;

	/// Whether or not to show the version in the Server header
	int showver;

	/// Max protocol version number
	int maxver;

	/// List of allowed server names (unused)
	char *servers;
	
	/// List of allowed schemes (unused)
	char *schemes;

	/// ErrorDoc to show when a hack attempt is caught
	char *errdoc;

	/// Command to execute in case of a hack attempt
	char *alarm_cfg;

	/// Per module configurations
	table *modcfg;
	
	/// linked list of WASap modules
	struct wasap_ext *wasap_exts;
	
} wasap_srvcfg;


/** WASap Directory configuration
 */
typedef struct {
	/// Allowed Methods
	/// corresponds with the M_* values
	int methods;
	
	/// Whether or not mod_wasap is activated or not
	int on;
	
	/// Max GET value length
	int maxgetval;
	
	/// Max POST value length
	int maxpostval;
	
	/// Max Header value length
	int maxheader;
	
	/// List of checks to perform
	table *checks;
	
	/// Per module configuation
	table *modcfg;
	
	/// Directory name
	char *dir;
	
	/// Command to execute in case of a hack attempt
	char *alarm_cfg;
	
} wasap_dircfg;

/** WASap Request Structure (portability layer)
 *
 */

typedef struct wasap_rec {
	/// Allocation Function for the Web Server
	void *(*alloc)(void *pool, int size);
	
	/// Allocation Pool or Context Pointer
	void *pool;
	
	/// Pointer to original Request or Context Pointer
	void *rec;
	
	/// WASap Server Configuration
	wasap_srvcfg *scfg;
	
	/// WASap Directory Configuration
	wasap_dircfg *dcfg;
	
	/// WASap tmp variable, used in check_headers()
	int tmp;
} wasap_rec;

char *trans_uri(pool *, const char *, int);

int check_proto(wasap_rec *, char *);
int check_scheme(wasap_rec *, char *);
int check_uri(wasap_rec *, char *);
